/* Print Motorola 68k instructions for GDB, the GNU debugger.
   Copyright 1986, 1987, 1989, 1991 Free Software Foundation, Inc.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdio.h>

struct m68k_opcode
{
  char *name;
  unsigned long opcode;
  unsigned long  match;
  char *args;
};

extern struct m68k_opcode m68k_opcodes[];
extern int numopcodes;
extern struct m68k_opcode *endop;

/* 68k instructions are never longer than this many bytes.  */
#define MAXLEN 22

char *reg_names[] = {
  "d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7",
  "a0", "a1", "a2", "a3", "a4", "a5", "fp", "sp"};

char *fpcr_names[] = { "", "fpiar", "fpsr", "fpiar/fpsr", "fpcr",
		     "fpiar/fpcr", "fpsr/fpcr", "fpiar-fpcr"};

static unsigned char *print_insn_arg ();
static unsigned char *insn_arg_size ();
static unsigned char *print_indexed ();
static unsigned char *indexed_size ();
static void print_base ();
static int fetch_arg ();
static void print_num();
void bad_format(char *);
void print_address(unsigned);
static FILE *out;

#define NEXTBYTE(p)  (p += 2, ((char *)p)[-1])

#define NEXTWORD(p)  \
  (p += 2, ((((char *)p)[-2]) << 8) + p[-1])

#define NEXTLONG(p)  \
  (p += 4, (((((p[-4] << 8) + p[-3]) << 8) + p[-2]) << 8) + p[-1])

#define NEXTSINGLE(p) \
  (p += 4, *((float *)(p - 4)))

#define NEXTDOUBLE(p) \
  (p += 8, *((double *)(p - 8)))

#define NEXTEXTEND(p) \
  (p += 12, 0.0)	/* Need a function to convert from extended to double
			   precision... */

#define NEXTPACKED(p) \
  (p += 12, 0.0)	/* Need a function to convert from packed to double
			   precision.   Actually, it's easier to print a
			   packed number than a double anyway, so maybe
			   there should be a special case to handle this... */

/* Print the m68k instruction at address MEMADDR in debugged memory,
   on STREAM.  Returns length of the instruction, in bytes.  */

typedef unsigned long CORE_ADDR;

int
print_m68k_insn (aout, memaddr, buffer)
     FILE *aout;
     CORE_ADDR memaddr;
     unsigned char *buffer;
{
  register int i;
  register unsigned char *p;
  register char *d;
  register int bestmask;
  int best;

  out = aout;
  bestmask = 0;
  best = -1;
  for (i = 0; i < numopcodes; i++)
    {
      register unsigned int opcode = m68k_opcodes[i].opcode;
      register unsigned int match = m68k_opcodes[i].match;
      if (((0xff & buffer[0] & (match >> 24)) == (0xff & (opcode >> 24)))
	  && ((0xff & buffer[1] & (match >> 16)) == (0xff & (opcode >> 16)))
	  && ((0xff & buffer[2] & (match >> 8)) == (0xff & (opcode >> 8)))
	  && ((0xff & buffer[3] & match) == (0xff & opcode)))
	{
	  /* Don't use for printout the variants of divul and divsl
	     that have the same register number in two places.
	     The more general variants will match instead.  */
	  for (d = m68k_opcodes[i].args; *d; d += 2)
	    if (d[1] == 'D')
	      break;

	  /* Don't use for printout the variants of most floating
	     point coprocessor instructions which use the same
	     register number in two places, as above. */
	  if (*d == 0)
	    for (d = m68k_opcodes[i].args; *d; d += 2)
	      if (d[1] == 't')
		break;

	  if (*d == 0 && match > bestmask)
	    {
	      best = i;
	      bestmask = match;
	    }
	}
    }

  /* Handle undefined instructions.  */
  if (best < 0)
    return 0;

  if( out != NULL )
    fprintf (out, "%s", m68k_opcodes[best].name);

  /* Point at first word of argument data,
     and at descriptor for first argument.  */
  p = buffer + 2;
  
  /* Why do this this way? -MelloN */
  for (d = m68k_opcodes[best].args; *d; d += 2)
    {
      if (d[0] == '#')
	{
	  if (d[1] == 'l' && p - buffer < 6)
	    p = buffer + 6;
	  else if (p - buffer < 4 && d[1] != 'C' && d[1] != '8' )
	    p = buffer + 4;
	}
      if (d[1] >= '1' && d[1] <= '3' && p - buffer < 4)
	p = buffer + 4;
      if (d[1] >= '4' && d[1] <= '6' && p - buffer < 6)
	p = buffer + 6;
      if ((d[0] == 'L' || d[0] == 'l') && d[1] == 'w' && p - buffer < 4)
	p = buffer + 4;
    }

  d = m68k_opcodes[best].args;

  if (*d && out != NULL)
    fprintf (out, " ");

  while (*d)
    {
      if( out != NULL )
        p = print_insn_arg (d, buffer, p, memaddr + p - buffer);
      else
	p = insn_arg_size(d, buffer, p);
      d += 2;
      if (out != NULL && *d && *(d - 2) != 'I' && *d != 'k')
	fprintf (out, ",");
    }
  return p - buffer;
}

static unsigned char *
print_insn_arg (d, buffer, p, addr)
     char *d;
     unsigned char *buffer;
     register unsigned char *p;
     CORE_ADDR addr;		/* PC for this arg to be relative to */
{
  register int val;
  register int place = d[1];
  int regno;
  register char *regname;
  register unsigned char *p1;
  register double flval;
  int flt_p;

  switch (*d)
    {
    case 'C':
      fprintf (out, "ccr");
      break;

    case 'S':
      fprintf (out, "sr");
      break;

    case 'U':
      fprintf (out, "usp");
      break;

    case 'J':
      {
	static struct { char *name; int value; } names[]
	  = {{"sfc", 0x000}, {"dfc", 0x001}, {"cacr", 0x002},
	     {"usp", 0x800}, {"vbr", 0x801}, {"caar", 0x802},
	     {"msp", 0x803}, {"isp", 0x804}};

	val = fetch_arg (buffer, place, 12);
	for (regno = sizeof names / sizeof names[0] - 1; regno >= 0; regno--)
	  if (names[regno].value == val)
	    {
	      fprintf (out, names[regno].name);
	      break;
	    }
	if (regno < 0)
	  fprintf (out, "0x%x", val);
      }
      break;

    case 'Q':
      val = fetch_arg (buffer, place, 3);
      /* 0 means 8, except for the bkpt instruction... */
      if (val == 0 && d[1] != 's')
	val = 8;
      fprintf (out, "#%d", val);
      break;

    case 'M':
      val = fetch_arg (buffer, place, 8);
      if (val & 0x80)
	val = val - 0x100;
      fprintf (out, "#%d", val);
      break;

    case 'T':
      val = fetch_arg (buffer, place, 4);
      fprintf (out, "#%d", val);
      break;

    case 'D':
      fprintf (out, "%s", reg_names[fetch_arg (buffer, place, 3)]);
      break;

    case 'A':
      fprintf (out, "%s", reg_names[fetch_arg (buffer, place, 3) + 010]);
      break;

    case 'R':
      fprintf (out, "%s", reg_names[fetch_arg (buffer, place, 4)]);
      break;

    case 'F':
      fprintf (out, "fp%d", fetch_arg (buffer, place, 3));
      break;

    case 'O':
      val = fetch_arg (buffer, place, 6);
      if (val & 0x20)
	fprintf (out, "%s", reg_names [val & 7]);
      else
	fprintf (out, "%d", val);
      break;

    case '+':
      fprintf (out, "%s@+", reg_names[fetch_arg (buffer, place, 3) + 8]);
      break;

    case '-':
      fprintf (out, "%s@-", reg_names[fetch_arg (buffer, place, 3) + 8]);
      break;

    case 'k':
      if (place == 'k')
	fprintf (out, "{%s}", reg_names[fetch_arg (buffer, place, 3)]);
      else if (place == 'C')
	{
	  val = fetch_arg (buffer, place, 7);
	  if ( val > 63 )		/* This is a signed constant. */
	    val -= 128;
	  fprintf (out, "{#%d}", val);
	}
      else
	bad_format(d);
      break;

    case '#':
    case '^':
      p1 = buffer + (*d == '#' ? 2 : 4);
      if (place == 's')
	val = fetch_arg (buffer, place, 4);
      else if (place == 'C')
	val = fetch_arg (buffer, place, 7);
      else if (place == '8')
	val = fetch_arg (buffer, place, 3);
      else if (place == '3')
	val = fetch_arg (buffer, place, 8);
      else if (place == 'b')
	val = NEXTBYTE (p1);
      else if (place == 'w')
	val = NEXTWORD (p1);
      else if (place == 'l')
	val = NEXTLONG (p1);
      else
	bad_format(d);
      fprintf(out, "#");
      print_num(val);
      break;

    case 'B':
      if (place == 'b')
	val = NEXTBYTE (p);
      else if (place == 'B')
	val = buffer[1];
      else if (place == 'w' || place == 'W')
	val = NEXTWORD (p);
      else if (place == 'l' || place == 'L')
	val = NEXTLONG (p);
      else if (place == 'g')
	{
	  val = ((char *)buffer)[1];
	  if (val == 0)
	    val = NEXTWORD (p);
	  else if (val == -1)
	    val = NEXTLONG (p);
	}
      else if (place == 'c')
	{
	  if (buffer[1] & 0x40)		/* If bit six is one, long offset */
	    val = NEXTLONG (p);
	  else
	    val = NEXTWORD (p);
	}
      else
	bad_format(d);

      print_address (addr + val);
      break;

    case 'd':
      val = NEXTWORD (p);
      fprintf (out, "%s@(", reg_names[fetch_arg (buffer, place, 3)]);
      print_num(val);
      fprintf (out, ")");
      break;

    case 's':
      fprintf (out, "%s", fpcr_names[fetch_arg (buffer, place, 3)]);
      break;

    case 'I':
      val = fetch_arg (buffer, 'd', 3);		  /* Get coprocessor ID... */
      if (val != 1)				/* Unusual coprocessor ID? */
	fprintf (out, "(cpid=%d) ", val);
      if (place == 'i')
	p += 2;			     /* Skip coprocessor extended operands */
      break;

    case '*':
    case '~':
    case '%':
    case ';':
    case '@':
    case '!':
    case '$':
    case '?':
    case '/':
    case '&':

      if (place == 'd')
	{
	  val = fetch_arg (buffer, 'x', 6);
	  val = ((val & 7) << 3) + ((val >> 3) & 7);
	}
      else
	val = fetch_arg (buffer, 's', 6);

      /* Get register number assuming address register.  */
      regno = (val & 7) + 8;
      regname = reg_names[regno];
      switch (val >> 3)
	{
	case 0:
	  fprintf (out, "%s", reg_names[val]);
	  break;

	case 1:
	  fprintf (out, "%s", regname);
	  break;

	case 2:
	  fprintf (out, "%s@", regname);
	  break;

	case 3:
	  fprintf (out, "%s@+", regname);
	  break;

	case 4:
	  fprintf (out, "%s@-", regname);
	  break;

	case 5:
	  val = NEXTWORD (p);
	  fprintf (out, "%s@(", regname);
	  print_num (val);
	  fprintf (out, ")");
	  break;

	case 6:
	  p = print_indexed (regno, p, addr);
	  break;

	case 7:
	  switch (val & 7)
	    {
	    case 0:
	      val = NEXTWORD (p);
	      fprintf (out, "@#");
	      print_address (val);
	      break;

	    case 1:
	      val = NEXTLONG (p);
	      fprintf (out, "@#");
	      print_address (val);
	      break;

	    case 2:
	      val = NEXTWORD (p);
	      print_address (addr + val);
	      break;

	    case 3:
	      p = print_indexed (-1, p, addr);
	      break;

	    case 4:
	      flt_p = 1;	/* Assume it's a float... */
	      switch( place )
	      {
		case 'b':
		  val = NEXTBYTE (p);
		  flt_p = 0;
		  break;

		case 'w':
		  val = NEXTWORD (p);
		  flt_p = 0;
		  break;

		case 'l':
		  val = NEXTLONG (p);
		  flt_p = 0;
		  break;

		case 'f':
		  flval = NEXTSINGLE(p);
		  break;

		case 'F':
		  flval = NEXTDOUBLE(p);
		  break;

		case 'x':
		  flval = NEXTEXTEND(p);
		  break;

		case 'p':
		  flval = NEXTPACKED(p);
		  break;

		default:
		  bad_format(d);
	      }
	      if ( flt_p )	/* Print a float? */
		fprintf (out, "#%g", flval);
	      else {
		fprintf (out, "#");
		print_num (val);
	      }
	      break;

	    default:
	      fprintf (out, "<invalid address mode %c%c>", ((val >> 3) & 7) + '0',
			(val & 7) + '0');
	    }
	}
      break;

    case 'L':
    case 'l':
	if (place == 'w')
	  {
	    char doneany;
	    p1 = buffer + 2;
	    val = NEXTWORD (p1);
	    /* Move the pointer ahead if this point is farther ahead
	       than the last.  */
	    p = p1 > p ? p1 : p;
	    if (val == 0)
	      {
		fprintf (out, "#0");
		break;
	      }
	    if (*d == 'l')
	      {
		register int newval = 0;
		for (regno = 0; regno < 16; ++regno)
		  if (val & (0x8000 >> regno))
		    newval |= 1 << regno;
		val = newval;
	      }
	    val &= 0xffff;
	    doneany = 0;
	    for (regno = 0; regno < 16; ++regno)
	      if (val & (1 << regno))
		{
		  int first_regno;
		  if (doneany)
		    fprintf (out, "/");
		  doneany = 1;
		  fprintf (out, "%s", reg_names[regno]);
		  first_regno = regno;
		  while (val & (1 << (regno + 1)))
		    ++regno;
		  if (regno > first_regno)
		    fprintf (out, "-%s", reg_names[regno]);
		}
	  }
	else if (place == '3')
	  {
	    /* `fmovem' insn.  */
	    char doneany;
	    val = fetch_arg (buffer, place, 8);
	    if (val == 0)
	      {
		fprintf (out, "#0");
		break;
	      }
	    if (*d == 'l')
	      {
		register int newval = 0;
		for (regno = 0; regno < 8; ++regno)
		  if (val & (0x80 >> regno))
		    newval |= 1 << regno;
		val = newval;
	      }
	    val &= 0xff;
	    doneany = 0;
	    for (regno = 0; regno < 8; ++regno)
	      if (val & (1 << regno))
		{
		  int first_regno;
		  if (doneany)
		    fprintf (out, "/");
		  doneany = 1;
		  fprintf (out, "fp%d", regno);
		  first_regno = regno;
		  while (val & (1 << (regno + 1)))
		    ++regno;
		  if (regno > first_regno)
		    fprintf (out, "-fp%d", regno);
		}
	  }
	else
	  bad_format(d);
      break;

    default:
      bad_format(d);
    }

  return (unsigned char *) p;
}

void
bad_format(d)
  char *d;
{
  fprintf(out, "?<%c%c>", d[0], d[1]);
}

/* Work out the size of an instruction argument. */
static unsigned char *
insn_arg_size (d, buffer, p)
     char *d;
     unsigned char *buffer;
     register unsigned char *p;
{
  register int val;
  register int place = d[1];
  register unsigned char *p1;

  switch (*d)
    {
    case 'B':
      if (place == 'b')
	NEXTBYTE (p);
      else if (place == 'w' || place == 'W')
	NEXTWORD (p);
      else if (place == 'l' || place == 'L')
	NEXTLONG (p);
      else if (place == 'g')
	{
	  val = ((char *)buffer)[1];
	  if (val == 0)
	    NEXTWORD (p);
	  else if (val == -1)
	    NEXTLONG (p);
	}
      else if (place == 'c')
	{
	  if (buffer[1] & 0x40)		/* If bit six is one, long offset */
	    NEXTLONG (p);
	  else
	    NEXTWORD (p);
	}

      break;

    case 'd':
      NEXTWORD (p);
      break;

    case 'I':
      if (place == 'i')
	p += 2;			     /* Skip coprocessor extended operands */
      break;

    case '*':
    case '~':
    case '%':
    case ';':
    case '@':
    case '!':
    case '$':
    case '?':
    case '/':
    case '&':

      if (place == 'd')
	{
	  val = fetch_arg (buffer, 'x', 6);
	  val = ((val & 7) << 3) + ((val >> 3) & 7);
	}
      else
	val = fetch_arg (buffer, 's', 6);

      switch (val >> 3)
	{
	case 5:
	  NEXTWORD (p);
	  break;

	case 6:
	  p = indexed_size (p);
	  break;

	case 7:
	  switch (val & 7)
	    {
	    case 0:
	      NEXTWORD (p);
	      break;

	    case 1:
	      NEXTLONG (p);
	      break;

	    case 2:
	      NEXTWORD (p);
	      break;

	    case 3:
	      p = indexed_size (p);
	      break;

	    case 4:
	      switch( place )
	      {
		case 'b':
		  NEXTBYTE (p);
		  break;

		case 'w':
		  NEXTWORD (p);
		  break;

		case 'l':
		  NEXTLONG (p);
		  break;

		case 'f':
		  NEXTSINGLE(p);
		  break;

		case 'F':
		  NEXTDOUBLE(p);
		  break;

		case 'x':
		  NEXTEXTEND(p);
		  break;

		case 'p':
		  NEXTPACKED(p);
		  break;

	      }
	      break;
	    }
	}
      break;

    case 'L':
    case 'l':
	if (place == 'w')
	  {
	    p1 = buffer + 2;
	    NEXTWORD (p1);
	    /* Move the pointer ahead if this point is farther ahead
	       than the last.  */
	    p = p1 > p ? p1 : p;
	  }
      break;
    }

  return (unsigned char *) p;
}

/* Fetch BITS bits from a position in the instruction specified by CODE.
   CODE is a "place to put an argument", or 'x' for a destination
   that is a general address (mode and register).
   BUFFER contains the instruction.  */

static int
fetch_arg (buffer, code, bits)
     unsigned char *buffer;
     char code;
     int bits;
{
  register int val;
  switch (code)
    {
    case 's':
      val = buffer[1];
      break;

    case 'd':			/* Destination, for register or quick.  */
      val = (buffer[0] << 8) + buffer[1];
      val >>= 9;
      break;

    case 'x':			/* Destination, for general arg */
      val = (buffer[0] << 8) + buffer[1];
      val >>= 6;
      break;

    case 'k':
      val = (buffer[3] >> 4);
      break;

    case 'C':
      val = buffer[3];
      break;

    case '1':
      val = (buffer[2] << 8) + buffer[3];
      val >>= 12;
      break;

    case '2':
      val = (buffer[2] << 8) + buffer[3];
      val >>= 6;
      break;

    case '3':
    case 'j':
      val = (buffer[2] << 8) + buffer[3];
      break;

    case '4':
      val = (buffer[4] << 8) + buffer[5];
      val >>= 12;
      break;

    case '5':
      val = (buffer[4] << 8) + buffer[5];
      val >>= 6;
      break;

    case '6':
      val = (buffer[4] << 8) + buffer[5];
      break;

    case '7':
      val = (buffer[2] << 8) + buffer[3];
      val >>= 7;
      break;
      
    case '8':
      val = (buffer[2] << 8) + buffer[3];
      val >>= 10;
      break;

    default:
      fprintf(out, "\nBad code value in fetch_arg: \\x%x\n", code);
    }

  return val & ((1 << bits) - 1);
}

/* Print an indexed argument.  The base register is BASEREG (-1 for pc).
   P points to extension word, in buffer.
   ADDR is the nominal core address of that extension word.  */

static unsigned char *
print_indexed (basereg, p, addr)
     int basereg;
     unsigned char *p;
     CORE_ADDR addr;
{
  register int word;
  static char *scales[] = {"", "*2", "*4", "*8"};
  register int base_disp;
  register int outer_disp;
  char *r, *q, buf[40];

  word = NEXTWORD (p);

  /* Generate the text for the index register.
     Where this will be output is not yet determined.  */
  r = buf;
  *r++ = '[';
  for( q = reg_names[(word >> 12) & 0xf]; *q != 0; )
    *r++ = *q++;
  *r++ = '.';
  *r++ = (word & 0x800) ? 'l' : 'w';
  for( q = scales[(word >> 9) & 3]; *q != 0; )
    *r++ = *q++;
  *r++ = ']';
  *r = 0;

  /* Handle the 68000 style of indexing.  */

  if ((word & 0x100) == 0)
    {
      print_base (basereg,
		  ((word & 0x80) ? word | 0xff00 : word & 0xff)
		  + ((basereg == -1) ? addr : 0));
      fprintf (out, buf);
      return p;
    }

  /* Handle the generalized kind.  */
  /* First, compute the displacement to add to the base register.  */

  if (word & 0200)
    basereg = -2;
  if (word & 0100)
    buf[0] = 0;
  base_disp = 0;
  switch ((word >> 4) & 3)
    {
    case 2:
      base_disp = NEXTWORD (p);
      break;
    case 3:
      base_disp = NEXTLONG (p);
    }
  if (basereg == -1)
    base_disp += addr;

  /* Handle single-level case (not indirect) */

  if ((word & 7) == 0)
    {
      print_base (basereg, base_disp);
      fprintf (out, buf);
      return p;
    }

  /* Two level.  Compute displacement to add after indirection.  */

  outer_disp = 0;
  switch (word & 3)
    {
    case 2:
      outer_disp = NEXTWORD (p);
      break;
    case 3:
      outer_disp = NEXTLONG (p);
    }

  print_num (outer_disp);
  fprintf (out, "(");
  print_base (basereg, base_disp);

  /* If postindexed, print the closeparen before the index.  */
  if (word & 4)
    fprintf (out, ")%s", buf);
  /* If preindexed, print the closeparen after the index.  */
  else
    fprintf (out, "%s)", buf);

  return p;
}

/* Work out the size of an indexed address. */
static unsigned char *
indexed_size (p)
     unsigned char *p;
{
  register int word;

  word = NEXTWORD (p);

  /* Handle the 68000 style of indexing.  */

  if ((word & 0x100) == 0)
    {
      return p;
    }

  /* Handle the generalized kind.  */
  /* First, compute the displacement to add to the base register.  */

  switch ((word >> 4) & 3)
    {
    case 2:
      NEXTWORD (p);
      break;
    case 3:
      NEXTLONG (p);
    }

  /* Handle single-level case (not indirect) */

  if ((word & 7) == 0)
    {
      return p;
    }

  /* Two level.  Compute displacement to add after indirection.  */

  switch (word & 3)
    {
    case 2:
      NEXTWORD (p);
      break;
    case 3:
      NEXTLONG (p);
    }

  return p;
}

/* Print a base register REGNO and displacement DISP, on STREAM.
   REGNO = -1 for pc, -2 for none (suppressed).  */

static void
print_base (regno, disp)
     int regno;
     int disp;
{
  if (regno == -2)
    print_num (disp);
  else if (regno == -1)
    fprintf (out, "0x%x", disp);
  else
    fprintf (out, "%d(%s)", disp, reg_names[regno]);
}

static void
print_num(val)
  int val;
{
  if( -32 <= val && val <= 32 )
    fprintf (out, "%d", val);
  else if( -0xFFFF <= val && val < 0 )
    fprintf (out, "-0x%x", -val);
  else
    fprintf (out, "0x%x", val);
}
